/*
	AOK Trigger Studio (See aokts.cpp for legal conditions.)
	WINDOWS VERSION.

	scen.h -- wraps a scenario file in class Scenario and defines the interface.

  Notes:
  This file's code is platform-independant albeit windows.h is included for the bitmap structs.
  'u' stands for an unknown member. (There are a lot of these.)
*/

#ifndef SCEN_H
#define SCEN_H
#pragma once

#include "datatypes.h"
#include "trigger.h"

//MAC USERS: Please replace this block of code with whatever you need.
#define WIN32_LEAN_AND_MEAN
#include <windows.h>	//for bitmaps, POINT, and RECT

#include <stdlib.h>	//for _MAX_FNAME and the like.
#include <time.h>		//for time_t typedef.

typedef struct _iobuf FILE;	//makes including <stdio.h> unnecessary

/* Options */

#define MAX_MAPSIZE		256	//maximum size of one side of a map
#define GAIA_INDEX		8	//index of GAIA players (may change)
#define PLAYER1_INDEX	0	//index of Player 1 (may change)

#define UNREAD(n)

/** Structs **/

struct AOKFile
{
	char name[_MAX_FNAME];
	SString data;
};

/* Map */

#pragma pack(push, 1)	//set packing alignment to work with terrain array

class Map	//wraps all data regarding the scenario map
{
public:
//in scenario

	long aitype;	//Map type for AI recognition. (see aitypes[])
	size_t x, y;	//Should always be equal, but are stored separately in the file.

	struct Terrain		//This is one tile of terrain.
	{
		Terrain();		//sets cnst and elev to default values.

		unsigned char cnst;	//terrain tile constant. (see esdata)
		unsigned char elev;	//terrain elevation. (0-8, I believe)

	} terrain[MAX_MAPSIZE][MAX_MAPSIZE];	//the terrain tile arrays

//internal

	Map();

	void reset();
	void read(FILE *in, bool ex);
	void write(FILE *out, bool ex);

	/*	readArea: copies terrain data from a buffer to the specified area */
	bool readArea(Buffer &from, const RECT &area);

	/*	writes a specified area of the map to a buffer

		area: inclusive area to copy

		returns: true if successful
	*/
	bool writeArea(Buffer &b, const RECT &area);
};

/* Units */

class Unit
{
public:
	Unit(UID id = 0);
	Unit(Buffer &from);

	float	y, x;		//stupid order
	float	u_float;	//always 1.0 (any use for this?)
	UID		ident;	//identifies unit for triggers, etc.
private:
	UCNST	cnst;		//constant (type) of unit
public:
	char	u_char;	//always 2
	float	rotate;	//in radians
	short	frame;
	UID		garrison;

	static size_t size;

	void read(FILE *in);

	const UnitLink *getType();
	const UnitLink *getType() const;
	void setType(const UnitLink *t);
//	int size() const;
	void toBuffer(Buffer &b) const;

private:
	const UnitLink *pType;
};

/* Players */

#define NUM_PLAYERS	0x10

enum AIModes
{
	AI_custom,
	AI_standard,
	AI_none
};

class Player
{
	void reset();
	bool read_aifile(FILE *in);	//returns false if out-of-memory

	bool read_units(FILE *in);
	void write_units(FILE *out);

	/*
		read_data3: Reads Player Data 3.

		in: The file to read from.
		view: Will read the camera floats to these 2 floats if nonzero.

		Returns false if ucount unrecognized.
	*/
	bool read_data3(FILE *in, float *view);
	void write_data3(FILE *out, int me, float *view);

public:
	Player();
	~Player();

	bool import_ai(const char *path);
	bool export_ai(const char *path);

	char name[30];	//256 bytes in file, but only allows 29 characters
	long stable;	//string table
	BOOL enable, human;
	long civ;

	char ai[_MAX_FNAME];
	char *aifile;	//memory-mapping of AI file for exporting
	char aimode;	//see enum AIModes

	/* Starting resource stockpiles.

		AOK: gold, wood, food, stone, orex, unknown
		SWGB: nova, carbon, food, ore, orex, unknown
	*/
	enum ResourceID
	{
		RES_gold, RES_wood, RES_food, RES_stone, RES_orex, RES_unknown
	};
	unsigned long resources[6];

	//disables
	UNREAD(long AlliedVictory);
	long dis_tech[60], dis_unit[60], dis_bldg[34];
	long ndis_t, ndis_u, ndis_b;

	long age;	//starting age

	float pop;	//EX-only
	float camera[2];	//initial camera position
	short u1, u2;
	bool avictory;	//allied victory
	long diplomacy[NUM_PLAYERS];
	long color;
	float ucount;

	SVector <Unit> units;

	static const char* names[NUM_PLAYERS];

	static int num_players;	//for diplomacy

	friend class Scenario;
};

#pragma pack(4)

struct Victory
{
	enum Modes
	{
		MODE_Standard, MODE_Conquest, MODE_Score, MODE_Time, MODE_Custom
	};

	long conq, z1, relics, z2, expl, z3;
	long all, mode, score, time;	//really bools, just stored as longs in-file
};

/*	MapCopyCache: caches data in-between Scenario::map_size() and Scenario::
	map_copy() so we don't have to do some loops twice. */
class MapCopyCache;

/* The Scenario Wrapper */

#define NUM_UNK 32

class Scenario
{

// Private structs
	struct AOKBMP
	{
		BITMAPFILEHEADER file_hdr;
		BITMAPINFOHEADER info_hdr;
		RGBQUAD *colors;
		char *image;

		AOKBMP();
		~AOKBMP();

		bool read(FILE *in);
		void write(FILE *out);
		bool toFile(const char *path);
		void reset();
	};

	const struct PerVersion
	{
		int messages_count;
		bool mstrings;
		int max_disables1, max_disables2;
	} *perversion;	//yes, I know that's a word. :-P

	static const PerVersion pv1_18;
	static const PerVersion pv1_21;
	static const PerVersion pv1_22;
	static const PerVersion pv1_30;

//	Internal Status Stuff (aka not in the SCX)
	bool mod_status;

	int read_data(const char *path);		//reads decompressed scenario data
	int write_data(const char *path);	//writes scenario data to be compressed

public:
	Scenario();
	~Scenario();
	static char StandardAI[];

	char msg[_MAX_FNAME + 100];	//contains any error message reported by a function.

	/* The scenario */

	enum Version
	{
		VER_AOE1		= 10,
		VER_AOE2		= 18,
		VER_AOE2TC	= 21,
		VER_SWGB    = 21
	} ver;	//set according to char version[].

//	Un-compressed Header
	struct _header
	{
		char version[5];	//Version string (+ 1 for NULL)
		long length;		//length of header (after this point)
		time_t timestamp;	//last save
		//left out instructions in header
		UNREAD(long pcount);

		void reset();
		bool read(FILE *scx);	//error return
		void write(FILE *scx, const SString *instr, long players);
	} header;

//	Compression starts here
	UID	next_uid;
	float	ver2;
	long	unknown;	//usually 1?
	char	origname[_MAX_FNAME];
	long	mstrings[NUM_MSGS];
	SString messages[NUM_MSGS];
	char	cinem[NUM_CINEM][0x20];	//pre-game, victory, lose, bkg
	long	bBitmap;
	AOKBMP bitmap;
	SString unk[NUM_UNK];	//probably part of PlayerData2
	Player players[NUM_PLAYERS];
	Victory vict;
	long all_techs;
	Map map;
	float	editor_pos[2];
	SVector <Trigger> triggers;	//Triggers
	SVector <unsigned long> t_order;

	long unk3, unk4;
	long cFiles;
	AOKFile *files;

	int open(const char *path, const char *dpath);
	int save(const char *path, const char *dpath, bool write);
	void reset();
	bool export_bmp(const char *path);
	//exFile: Exports all (index of -1) or one file to the specified directory
	bool exFile(const char *directory, long index);

	/*
		clean_triggers: cleans trigger list of unused triggers

		Note: Necessary because the UI only deletes the t_order entry, not the trigger itself.
	*/
	void clean_triggers();

	/*
		insert_trigger: Inserts a copy of a trigger into the trigger list after a specified
		trigger.

		t: The trigger to copy.
		after: The index of the preceding trigger.

		Returns the index of the inserted trigger in the trigger vector.

		Note: This both appends the trigger to the trigger list and then inserts it into
		t_order.
	*/
	index_t insert_trigger(Trigger *t, index_t after);

	/*
		find_trigger: returns a trigger item in the t_order array
	*/
	unsigned long *find_trigger(unsigned long trigger);

	/*	map_size: returns size of memory needed to copy rectangle.
	
		We need to do this since the app has to allocate the memory using Win API
		functions for clipboard.
	*/
	int map_size(const RECT &source) const;

	/*	map_copy: copies terrain and units from one area of the map to another.
	
		WARNING: This function assumes source is a valid rectangle, ie. (top < bottom &&
		left < right).

		Note: The returned MapCopyCache object is deleted in map_copy().
	*/
	int map_size(const RECT &source, MapCopyCache *&mcc);

	/*	map_copy: copies terrain and units from a specified area of the map
		to a MapCopyData structure.

		Returns: Error code

		Note: MapCopyCache is deleted by this function.
	*/
	AOKTS_ERROR map_copy(Buffer &to, const MapCopyCache *mcc);

	/*	map_paste: "pastes" terrain and units from a buffer to a specified point.

		Note: The buffer should have been created by map_copy.
	*/
	AOKTS_ERROR map_paste(const POINT &to, Buffer &from);

	int getPlayerCount();

	bool needsave()	{ return mod_status; }

	bool isExpansion();
};

extern class Scenario scen;

#undef UNREAD

#pragma pack(pop)	//restore default packing alignment

#endif //SCEN_H
